"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var URI = require("urijs");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeNames_1 = require("./TypeNames");
var TypeNames_2 = require("./TypeNames");
var TypeAttributes_1 = require("./TypeAttributes");
var PathElementKind;
(function (PathElementKind) {
    PathElementKind[PathElementKind["Root"] = 0] = "Root";
    PathElementKind[PathElementKind["KeyOrIndex"] = 1] = "KeyOrIndex";
    PathElementKind[PathElementKind["Type"] = 2] = "Type";
    PathElementKind[PathElementKind["Object"] = 3] = "Object";
})(PathElementKind = exports.PathElementKind || (exports.PathElementKind = {}));
function keyOrIndex(pe) {
    if (pe.kind !== PathElementKind.KeyOrIndex)
        return undefined;
    return pe.key;
}
function pathElementEquals(a, b) {
    if (a.kind !== b.kind)
        return false;
    switch (a.kind) {
        case PathElementKind.Type:
            return a.index === b.index;
        case PathElementKind.KeyOrIndex:
            return a.key === b.key;
        default:
            return true;
    }
}
function checkJSONSchema(x) {
    if (typeof x === "boolean")
        return x;
    if (Array.isArray(x))
        return Support_1.panic("An array is not a valid JSON Schema");
    if (x === null)
        return Support_1.panic("null is not a valid JSON Schema");
    if (typeof x !== "object")
        return Support_1.panic("Only booleans and objects can be valid JSON Schemas");
    return x;
}
exports.checkJSONSchema = checkJSONSchema;
var numberRegexp = new RegExp("^[0-9]+$");
var Ref = /** @class */ (function () {
    function Ref(addressURI, path) {
        this.path = path;
        if (addressURI !== undefined) {
            Support_1.assert(addressURI.fragment() === "", "Ref URI with fragment is not allowed: " + addressURI.toString());
            this.addressURI = addressURI.clone().normalize();
        }
        else {
            this.addressURI = undefined;
        }
    }
    Ref.root = function (address) {
        var uri = new URI(address);
        return new Ref(uri, immutable_1.List());
    };
    Ref.parsePath = function (path) {
        var elements = [];
        if (path.startsWith("/")) {
            elements.push({ kind: PathElementKind.Root });
            path = path.substr(1);
        }
        if (path !== "") {
            var parts = path.split("/");
            for (var i = 0; i < parts.length; i++) {
                elements.push({ kind: PathElementKind.KeyOrIndex, key: parts[i] });
            }
        }
        return immutable_1.List(elements);
    };
    Ref.parse = function (ref) {
        if (typeof ref !== "string") {
            return Support_1.panic("$ref must be a string");
        }
        var uri = new URI(ref);
        var path = uri.fragment();
        uri.fragment("");
        var elements = Ref.parsePath(path);
        return new Ref(uri, elements);
    };
    Object.defineProperty(Ref.prototype, "hasAddress", {
        get: function () {
            return this.addressURI !== undefined;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Ref.prototype, "address", {
        get: function () {
            return Support_1.defined(this.addressURI).toString();
        },
        enumerable: true,
        configurable: true
    });
    Ref.prototype.pushElement = function (pe) {
        return new Ref(this.addressURI, this.path.push(pe));
    };
    Ref.prototype.push = function () {
        var keys = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys[_i] = arguments[_i];
        }
        var ref = this;
        for (var _a = 0, keys_1 = keys; _a < keys_1.length; _a++) {
            var key = keys_1[_a];
            ref = ref.pushElement({ kind: PathElementKind.KeyOrIndex, key: key });
        }
        return ref;
    };
    Ref.prototype.pushObject = function () {
        return this.pushElement({ kind: PathElementKind.Object });
    };
    Ref.prototype.pushType = function (index) {
        return this.pushElement({ kind: PathElementKind.Type, index: index });
    };
    Ref.prototype.resolveAgainst = function (base) {
        var addressURI = this.addressURI;
        if (base !== undefined && base.addressURI !== undefined) {
            addressURI = addressURI === undefined ? base.addressURI : addressURI.absoluteTo(base.addressURI);
        }
        return new Ref(addressURI, this.path);
    };
    Object.defineProperty(Ref.prototype, "name", {
        get: function () {
            var path = this.path;
            for (;;) {
                var e = path.last();
                if (e === undefined || e.kind === PathElementKind.Root) {
                    var name = this.addressURI !== undefined ? this.addressURI.filename() : "";
                    var suffix = this.addressURI !== undefined ? this.addressURI.suffix() : "";
                    if (name.length > suffix.length + 1) {
                        name = name.substr(0, name.length - suffix.length - 1);
                    }
                    if (name === "") {
                        return "Something";
                    }
                    return name;
                }
                switch (e.kind) {
                    case PathElementKind.KeyOrIndex:
                        if (e.key.match(numberRegexp) !== null) {
                            return e.key;
                        }
                        break;
                    case PathElementKind.Type:
                    case PathElementKind.Object:
                        return Support_1.panic("We shouldn't try to get the name of Type or Object refs");
                    default:
                        return Support_1.assertNever(e);
                }
                path = path.pop();
            }
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Ref.prototype, "definitionName", {
        get: function () {
            var pe = this.path.get(-2);
            if (pe === undefined)
                return undefined;
            if (keyOrIndex(pe) === "definitions")
                return keyOrIndex(Support_1.defined(this.path.last()));
            return undefined;
        },
        enumerable: true,
        configurable: true
    });
    Ref.prototype.toString = function () {
        function elementToString(e) {
            switch (e.kind) {
                case PathElementKind.Root:
                    return "/";
                case PathElementKind.Type:
                    return "type/" + e.index.toString();
                case PathElementKind.Object:
                    return "object";
                case PathElementKind.KeyOrIndex:
                    return e.key;
                default:
                    return Support_1.assertNever(e);
            }
        }
        var address = this.addressURI === undefined ? "" : this.addressURI.toString();
        return address + "#" + this.path.map(elementToString).join("/");
    };
    Ref.prototype.lookupRef = function (root) {
        function lookup(local, path) {
            var first = path.first();
            if (first === undefined) {
                return checkJSONSchema(local);
            }
            var rest = path.rest();
            switch (first.kind) {
                case PathElementKind.Root:
                    return lookup(root, rest);
                case PathElementKind.KeyOrIndex:
                    if (Array.isArray(local)) {
                        return lookup(local[parseInt(first.key, 10)], rest);
                    }
                    else {
                        return lookup(Support_1.checkStringMap(local)[first.key], rest);
                    }
                case PathElementKind.Type:
                    return Support_1.panic('Cannot look up path that indexes "type"');
                case PathElementKind.Object:
                    return Support_1.panic('Cannot look up path that indexes "object"');
                default:
                    return Support_1.assertNever(first);
            }
        }
        return lookup(root, this.path);
    };
    Ref.prototype.equals = function (other) {
        if (!(other instanceof Ref))
            return false;
        if (this.addressURI !== undefined && other.addressURI !== undefined) {
            if (!this.addressURI.equals(other.addressURI))
                return false;
        }
        else {
            if ((this.addressURI === undefined) !== (other.addressURI === undefined))
                return false;
        }
        if (this.path.size !== other.path.size)
            return false;
        return this.path.zipWith(pathElementEquals, other.path).every(function (x) { return x; });
    };
    Ref.prototype.hashCode = function () {
        var acc = immutable_1.hash(this.addressURI !== undefined ? this.addressURI.toString() : undefined);
        this.path.forEach(function (pe) {
            acc = Support_1.addHashCode(acc, pe.kind);
            switch (pe.kind) {
                case PathElementKind.Type:
                    acc = Support_1.addHashCode(acc, pe.index);
                    break;
                case PathElementKind.KeyOrIndex:
                    acc = Support_1.addHashCode(acc, immutable_1.hash(pe.key));
                    break;
                default:
                    break;
            }
        });
        return acc;
    };
    return Ref;
}());
exports.Ref = Ref;
var Location = /** @class */ (function () {
    function Location(canonicalRef, virtualRef) {
        this.canonicalRef = canonicalRef;
        this.virtualRef = virtualRef !== undefined ? virtualRef : canonicalRef;
    }
    Location.prototype.updateWithID = function (id) {
        if (typeof id !== "string")
            return this;
        // FIXME: This is incorrect.  If the parsed ref doesn't have an address, the
        // current virtual one's must be used.  The canonizer must do this, too.
        return new Location(this.canonicalRef, Ref.parse(id).resolveAgainst(this.virtualRef));
    };
    Location.prototype.push = function () {
        var keys = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            keys[_i] = arguments[_i];
        }
        return new Location((_a = this.canonicalRef).push.apply(_a, keys), (_b = this.virtualRef).push.apply(_b, keys));
        var _a, _b;
    };
    Location.prototype.pushObject = function () {
        return new Location(this.canonicalRef.pushObject(), this.virtualRef.pushObject());
    };
    Location.prototype.pushType = function (index) {
        return new Location(this.canonicalRef.pushType(index), this.virtualRef.pushType(index));
    };
    Location.prototype.toString = function () {
        return this.virtualRef.toString() + " (" + this.canonicalRef.toString() + ")";
    };
    return Location;
}());
var Canonizer = /** @class */ (function () {
    function Canonizer() {
        this._map = immutable_1.Map();
        this._schemaAddressesAdded = immutable_1.Set();
    }
    Canonizer.prototype.addID = function (mapped, loc) {
        var ref = Ref.parse(mapped).resolveAgainst(loc.virtualRef);
        Support_1.assert(ref.hasAddress, "$id must have an address");
        this._map = this._map.set(ref, loc.canonicalRef);
    };
    Canonizer.prototype.addIDs = function (schema, loc) {
        if (schema === null)
            return;
        if (Array.isArray(schema)) {
            for (var i = 0; i < schema.length; i++) {
                this.addIDs(schema[i], loc.push(i.toString()));
            }
            return;
        }
        if (typeof schema !== "object") {
            return;
        }
        var maybeID = schema["$id"];
        if (typeof maybeID === "string") {
            this.addID(maybeID, loc);
            loc = loc.updateWithID(maybeID);
        }
        for (var _i = 0, _a = Object.getOwnPropertyNames(schema); _i < _a.length; _i++) {
            var property = _a[_i];
            this.addIDs(schema[property], loc.push(property));
        }
    };
    Canonizer.prototype.addSchema = function (schema, address) {
        if (this._schemaAddressesAdded.has(address))
            return;
        this.addIDs(schema, new Location(Ref.root(address)));
        this._schemaAddressesAdded = this._schemaAddressesAdded.add(address);
    };
    // Returns: Canonical ref, full virtual ref
    Canonizer.prototype.canonize = function (virtualBase, ref) {
        var fullVirtual = ref.resolveAgainst(virtualBase);
        var virtual = fullVirtual;
        var relative = immutable_1.List();
        for (;;) {
            var maybeCanonical = this._map.get(virtual);
            if (maybeCanonical !== undefined) {
                return [new Ref(maybeCanonical.addressURI, maybeCanonical.path.concat(relative)), fullVirtual];
            }
            var last = virtual.path.last();
            if (last === undefined) {
                // We've exhausted our options - it's not a mapped ref.
                return [fullVirtual, fullVirtual];
            }
            if (last.kind !== PathElementKind.Root) {
                relative = relative.unshift(last);
            }
            virtual = new Ref(virtual.addressURI, virtual.path.pop());
        }
    };
    return Canonizer;
}());
function checkStringArray(arr) {
    if (!Array.isArray(arr)) {
        return Support_1.panic("Expected a string array, but got " + arr);
    }
    for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
        var e = arr_1[_i];
        if (typeof e !== "string") {
            return Support_1.panic("Expected string, but got " + e);
        }
    }
    return arr;
}
function makeAttributes(schema, loc, attributes) {
    var maybeDescription = schema.description;
    if (typeof maybeDescription === "string") {
        attributes = TypeAttributes_1.descriptionTypeAttributeKind.setInAttributes(attributes, immutable_1.OrderedSet([maybeDescription]));
    }
    return TypeNames_2.modifyTypeNames(attributes, function (maybeTypeNames) {
        var typeNames = Support_1.defined(maybeTypeNames);
        if (!typeNames.areInferred) {
            return typeNames;
        }
        var title = schema.title;
        if (typeof title !== "string") {
            title = loc.canonicalRef.definitionName;
        }
        if (typeof title === "string") {
            return new TypeNames_1.TypeNames(immutable_1.OrderedSet([title]), immutable_1.OrderedSet(), false);
        }
        else {
            return typeNames.makeInferred();
        }
    });
}
function checkTypeList(typeOrTypes) {
    if (typeof typeOrTypes === "string") {
        return immutable_1.OrderedSet([typeOrTypes]);
    }
    else if (Array.isArray(typeOrTypes)) {
        var arr = [];
        for (var _i = 0, typeOrTypes_1 = typeOrTypes; _i < typeOrTypes_1.length; _i++) {
            var t = typeOrTypes_1[_i];
            if (typeof t !== "string") {
                return Support_1.panic("element of type is not a string: " + t);
            }
            arr.push(t);
        }
        var set = immutable_1.OrderedSet(arr);
        Support_1.assert(!set.isEmpty(), "JSON Schema must specify at least one type");
        return set;
    }
    else {
        return Support_1.panic("type is neither a string or array of strings: " + typeOrTypes);
    }
}
function addTypesInSchema(typeBuilder, store, references) {
    return __awaiter(this, void 0, void 0, function () {
        var _this = this;
        function resolveVirtualRef(base, virtualRef) {
            return __awaiter(this, void 0, void 0, function () {
                var _a, canonical, fullVirtual, schema;
                return __generator(this, function (_b) {
                    switch (_b.label) {
                        case 0:
                            _a = canonizer.canonize(base !== undefined ? base.virtualRef : undefined, virtualRef), canonical = _a[0], fullVirtual = _a[1];
                            Support_1.assert(canonical.hasAddress, "Canonical ref can't be resolved without an address");
                            return [4 /*yield*/, store.get(canonical.address)];
                        case 1:
                            schema = _b.sent();
                            canonizer.addSchema(schema, canonical.address);
                            return [2 /*return*/, [canonical.lookupRef(schema), new Location(canonical, fullVirtual)]];
                    }
                });
            });
        }
        function setTypeForLocation(loc, t) {
            return __awaiter(this, void 0, void 0, function () {
                var maybeRef;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, typeForCanonicalRef.get(loc.canonicalRef)];
                        case 1:
                            maybeRef = _a.sent();
                            if (maybeRef !== undefined) {
                                Support_1.assert(maybeRef === t, "Trying to set path again to different type");
                            }
                            typeForCanonicalRef = typeForCanonicalRef.set(loc.canonicalRef, t);
                            return [2 /*return*/];
                    }
                });
            });
        }
        function makeClass(loc, attributes, properties, requiredArray, additionalPropertiesType) {
            return __awaiter(this, void 0, void 0, function () {
                var _this = this;
                var required, propertiesMap, propertyDescriptions, props, additionalRequired, t_1, additionalProps;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            required = immutable_1.OrderedSet(requiredArray);
                            propertiesMap = immutable_1.Map(properties);
                            propertyDescriptions = propertiesMap
                                .map(function (propSchema) {
                                if (typeof propSchema === "object") {
                                    var desc = propSchema.description;
                                    if (typeof desc === "string") {
                                        return immutable_1.OrderedSet([desc]);
                                    }
                                }
                                return undefined;
                            })
                                .filter(function (v) { return v !== undefined; });
                            if (!propertyDescriptions.isEmpty()) {
                                attributes = TypeAttributes_1.propertyDescriptionsTypeAttributeKind.setInAttributes(attributes, propertyDescriptions);
                            }
                            return [4 /*yield*/, Support_1.mapSync(propertiesMap, function (propSchema, propName) { return __awaiter(_this, void 0, void 0, function () {
                                    var t, isOptional;
                                    return __generator(this, function (_a) {
                                        switch (_a.label) {
                                            case 0: return [4 /*yield*/, toType(Support_1.checkStringMap(propSchema), loc.push("properties", propName), TypeNames_2.makeNamesTypeAttributes(pluralize.singular(propName), true))];
                                            case 1:
                                                t = _a.sent();
                                                isOptional = !required.has(propName);
                                                return [2 /*return*/, new Type_1.ClassProperty(t, isOptional)];
                                        }
                                    });
                                }); })];
                        case 1:
                            props = (_a.sent()).toOrderedMap();
                            additionalRequired = required.subtract(props.keySeq());
                            if (!additionalRequired.isEmpty()) {
                                if (additionalPropertiesType === false) {
                                    return [2 /*return*/, Support_1.panic("Can't have non-specified required properties but forbidden additionalTypes")];
                                }
                                if (additionalPropertiesType === true) {
                                    t_1 = typeBuilder.getPrimitiveType("any");
                                }
                                else {
                                    t_1 = additionalPropertiesType;
                                }
                                additionalProps = additionalRequired.toOrderedMap().map(function (_name) { return new Type_1.ClassProperty(t_1, true); });
                                props = props.merge(additionalProps);
                            }
                            return [2 /*return*/, typeBuilder.getUniqueClassType(attributes, true, props)];
                    }
                });
            });
        }
        function makeMap(loc, typeAttributes, additional) {
            return __awaiter(this, void 0, void 0, function () {
                var valuesType;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            loc = loc.push("additionalProperties");
                            return [4 /*yield*/, toType(additional, loc, TypeNames_2.singularizeTypeNames(typeAttributes))];
                        case 1:
                            valuesType = _a.sent();
                            return [2 /*return*/, [typeBuilder.getMapType(valuesType), valuesType]];
                    }
                });
            });
        }
        function convertToType(schema, loc, typeAttributes) {
            return __awaiter(this, void 0, void 0, function () {
                function makeStringType() {
                    if (schema.format !== undefined) {
                        switch (schema.format) {
                            case "date":
                                return typeBuilder.getPrimitiveType("date");
                            case "time":
                                return typeBuilder.getPrimitiveType("time");
                            case "date-time":
                                return typeBuilder.getPrimitiveType("date-time");
                            default:
                                // FIXME: Output a warning here instead to indicate that
                                // the format is uninterpreted.
                                return typeBuilder.getStringType(inferredAttributes, undefined);
                        }
                    }
                    return typeBuilder.getStringType(inferredAttributes, undefined);
                }
                function makeArrayType() {
                    return __awaiter(this, void 0, void 0, function () {
                        var _a, _b;
                        return __generator(this, function (_c) {
                            switch (_c.label) {
                                case 0:
                                    if (!(schema.items !== undefined)) return [3 /*break*/, 2];
                                    loc = loc.push("items");
                                    _b = (_a = typeBuilder).getArrayType;
                                    return [4 /*yield*/, toType(Support_1.checkStringMap(schema.items), loc, TypeNames_2.singularizeTypeNames(typeAttributes))];
                                case 1: return [2 /*return*/, _b.apply(_a, [_c.sent()])];
                                case 2: return [2 /*return*/, typeBuilder.getArrayType(typeBuilder.getPrimitiveType("any"))];
                            }
                        });
                    });
                }
                function makeObjectTypes() {
                    return __awaiter(this, void 0, void 0, function () {
                        var required, typesInUnion, additionalPropertiesType, additional, _a, _b, _c, mapType, valuesType, _d, _e, _f, _g;
                        return __generator(this, function (_h) {
                            switch (_h.label) {
                                case 0:
                                    if (schema.required === undefined) {
                                        required = [];
                                    }
                                    else {
                                        required = checkStringArray(schema.required);
                                    }
                                    typesInUnion = [];
                                    additionalPropertiesType = true;
                                    if (!(schema.additionalProperties !== undefined)) return [3 /*break*/, 6];
                                    additional = schema.additionalProperties;
                                    if (!(additional === false)) return [3 /*break*/, 3];
                                    if (!(schema.properties === undefined)) return [3 /*break*/, 2];
                                    _b = (_a = typesInUnion).push;
                                    return [4 /*yield*/, makeClass(loc, inferredAttributes, {}, required, false)];
                                case 1:
                                    _b.apply(_a, [_h.sent()]);
                                    _h.label = 2;
                                case 2:
                                    additionalPropertiesType = false;
                                    return [3 /*break*/, 6];
                                case 3:
                                    if (!(typeof additional === "object")) return [3 /*break*/, 6];
                                    return [4 /*yield*/, makeMap(loc, inferredAttributes, Support_1.checkStringMap(additional))];
                                case 4:
                                    _c = _h.sent(), mapType = _c[0], valuesType = _c[1];
                                    _e = (_d = typesInUnion).push;
                                    return [4 /*yield*/, mapType];
                                case 5:
                                    _e.apply(_d, [_h.sent()]);
                                    additionalPropertiesType = valuesType;
                                    _h.label = 6;
                                case 6:
                                    if (!(schema.properties !== undefined)) return [3 /*break*/, 8];
                                    _g = (_f = typesInUnion).push;
                                    return [4 /*yield*/, makeClass(loc, inferredAttributes, Support_1.checkStringMap(schema.properties), required, additionalPropertiesType)];
                                case 7:
                                    _g.apply(_f, [_h.sent()]);
                                    _h.label = 8;
                                case 8:
                                    if (typesInUnion.length === 0) {
                                        typesInUnion.push(typeBuilder.getMapType(typeBuilder.getPrimitiveType("any")));
                                    }
                                    return [2 /*return*/, typesInUnion];
                            }
                        });
                    });
                }
                function makeTypesFromCases(cases, kind) {
                    return __awaiter(this, void 0, void 0, function () {
                        var _this = this;
                        return __generator(this, function (_a) {
                            switch (_a.label) {
                                case 0:
                                    if (!Array.isArray(cases)) {
                                        return [2 /*return*/, Support_1.panic("Cases are not an array: " + cases)];
                                    }
                                    return [4 /*yield*/, Support_1.mapSync(cases, function (t, index) { return __awaiter(_this, void 0, void 0, function () {
                                            return __generator(this, function (_a) {
                                                switch (_a.label) {
                                                    case 0: return [4 /*yield*/, toType(Support_1.checkStringMap(t), loc.push(kind, index.toString()), TypeAttributes_1.makeTypeAttributesInferred(typeAttributes))];
                                                    case 1: return [2 /*return*/, _a.sent()];
                                                }
                                            });
                                        }); })];
                                case 1: 
                                // FIXME: This cast shouldn't be necessary, but TypeScript forces our hand.
                                return [2 /*return*/, _a.sent()];
                            }
                        });
                    });
                }
                function convertOneOrAnyOf(cases, kind) {
                    return __awaiter(this, void 0, void 0, function () {
                        var unionType, _a, _b, _c, _d;
                        return __generator(this, function (_e) {
                            switch (_e.label) {
                                case 0:
                                    unionType = typeBuilder.getUniqueUnionType(TypeAttributes_1.makeTypeAttributesInferred(typeAttributes), undefined);
                                    _b = (_a = typeBuilder).setSetOperationMembers;
                                    _c = [unionType];
                                    _d = immutable_1.OrderedSet;
                                    return [4 /*yield*/, makeTypesFromCases(cases, kind)];
                                case 1:
                                    _b.apply(_a, _c.concat([_d.apply(void 0, [_e.sent()])]));
                                    return [2 /*return*/, unionType];
                            }
                        });
                    });
                }
                function includePrimitiveType(name) {
                    if (typeSet !== undefined && !typeSet.has(name)) {
                        return false;
                    }
                    if (enumArray !== undefined) {
                        var predicate = void 0;
                        switch (name) {
                            case "null":
                                predicate = function (x) { return x === null; };
                                break;
                            case "integer":
                                predicate = function (x) { return typeof x === "number" && x === Math.floor(x); };
                                break;
                            default:
                                predicate = function (x) { return typeof x === name; };
                                break;
                        }
                        return enumArray.find(predicate) !== undefined;
                    }
                    return true;
                }
                var inferredAttributes, enumArray, typeSet, includeObject, includeArray, needStringEnum, needUnion, intersectionType, types, unionTypes, _i, _a, _b, name, kind, cases, _c, _d, _e, _f, _g, virtualRef, _h, target, newLoc_1, attributes, _j, _k, _l, _m, _o, _p, _q, _r, _s;
                return __generator(this, function (_t) {
                    switch (_t.label) {
                        case 0:
                            typeAttributes = makeAttributes(schema, loc, typeAttributes);
                            inferredAttributes = TypeAttributes_1.makeTypeAttributesInferred(typeAttributes);
                            enumArray = Array.isArray(schema.enum) ? schema.enum : undefined;
                            typeSet = schema.type !== undefined ? checkTypeList(schema.type) : undefined;
                            includeObject = enumArray === undefined && (typeSet === undefined || typeSet.has("object"));
                            includeArray = enumArray === undefined && (typeSet === undefined || typeSet.has("array"));
                            needStringEnum = includePrimitiveType("string") && enumArray !== undefined && enumArray.find(function (x) { return typeof x === "string"; }) !== undefined;
                            needUnion = typeSet !== undefined || schema.properties !== undefined || schema.additionalProperties !== undefined || schema.items !== undefined || enumArray !== undefined;
                            intersectionType = typeBuilder.getUniqueIntersectionType(typeAttributes, undefined);
                            return [4 /*yield*/, setTypeForLocation(loc, intersectionType)];
                        case 1:
                            _t.sent();
                            types = [];
                            if (!needUnion) return [3 /*break*/, 6];
                            unionTypes = [];
                            for (_i = 0, _a = [["null", "null"], ["number", "double"], ["integer", "integer"], ["boolean", "bool"]]; _i < _a.length; _i++) {
                                _b = _a[_i], name = _b[0], kind = _b[1];
                                if (!includePrimitiveType(name))
                                    continue;
                                unionTypes.push(typeBuilder.getPrimitiveType(kind));
                            }
                            if (needStringEnum) {
                                cases = enumArray;
                                cases = cases.filter(function (x) { return typeof x === "string"; });
                                unionTypes.push(typeBuilder.getEnumType(inferredAttributes, immutable_1.OrderedSet(cases)));
                            }
                            else if (includePrimitiveType("string")) {
                                unionTypes.push(makeStringType());
                            }
                            if (!includeArray) return [3 /*break*/, 3];
                            _d = (_c = unionTypes).push;
                            return [4 /*yield*/, makeArrayType()];
                        case 2:
                            _d.apply(_c, [_t.sent()]);
                            _t.label = 3;
                        case 3:
                            if (!includeObject) return [3 /*break*/, 5];
                            _f = (_e = unionTypes.push).apply;
                            _g = [unionTypes];
                            return [4 /*yield*/, makeObjectTypes()];
                        case 4:
                            _f.apply(_e, _g.concat([_t.sent()]));
                            _t.label = 5;
                        case 5:
                            types.push(typeBuilder.getUniqueUnionType(inferredAttributes, immutable_1.OrderedSet(unionTypes)));
                            _t.label = 6;
                        case 6:
                            if (!(schema.$ref !== undefined)) return [3 /*break*/, 9];
                            virtualRef = Ref.parse(schema.$ref);
                            return [4 /*yield*/, resolveVirtualRef(loc, virtualRef)];
                        case 7:
                            _h = _t.sent(), target = _h[0], newLoc_1 = _h[1];
                            attributes = TypeNames_2.modifyTypeNames(typeAttributes, function (tn) {
                                if (!Support_1.defined(tn).areInferred)
                                    return tn;
                                return new TypeNames_1.TypeNames(immutable_1.OrderedSet([newLoc_1.canonicalRef.name]), immutable_1.OrderedSet(), true);
                            });
                            _k = (_j = types).push;
                            return [4 /*yield*/, toType(target, newLoc_1, attributes)];
                        case 8:
                            _k.apply(_j, [_t.sent()]);
                            _t.label = 9;
                        case 9:
                            if (!(schema.allOf !== undefined)) return [3 /*break*/, 11];
                            _m = (_l = types.push).apply;
                            _o = [types];
                            return [4 /*yield*/, makeTypesFromCases(schema.allOf, "allOf")];
                        case 10:
                            _m.apply(_l, _o.concat([(_t.sent())]));
                            _t.label = 11;
                        case 11:
                            if (!(schema.oneOf !== undefined)) return [3 /*break*/, 13];
                            _q = (_p = types).push;
                            return [4 /*yield*/, convertOneOrAnyOf(schema.oneOf, "oneOf")];
                        case 12:
                            _q.apply(_p, [_t.sent()]);
                            _t.label = 13;
                        case 13:
                            if (!(schema.anyOf !== undefined)) return [3 /*break*/, 15];
                            _s = (_r = types).push;
                            return [4 /*yield*/, convertOneOrAnyOf(schema.anyOf, "anyOf")];
                        case 14:
                            _s.apply(_r, [_t.sent()]);
                            _t.label = 15;
                        case 15:
                            typeBuilder.setSetOperationMembers(intersectionType, immutable_1.OrderedSet(types));
                            return [2 /*return*/, intersectionType];
                    }
                });
            });
        }
        function toType(schema, loc, typeAttributes) {
            return __awaiter(this, void 0, void 0, function () {
                var maybeType, result;
                return __generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            maybeType = typeForCanonicalRef.get(loc.canonicalRef);
                            if (maybeType !== undefined) {
                                return [2 /*return*/, maybeType];
                            }
                            if (!(typeof schema === "boolean")) return [3 /*break*/, 1];
                            // FIXME: Empty union.  We'd have to check that it's supported everywhere,
                            // in particular in union flattening.
                            Support_1.assert(schema === true, 'Schema "false" is not supported');
                            result = typeBuilder.getPrimitiveType("any");
                            return [3 /*break*/, 3];
                        case 1:
                            loc = loc.updateWithID(schema["$id"]);
                            return [4 /*yield*/, convertToType(schema, loc, typeAttributes)];
                        case 2:
                            result = _a.sent();
                            _a.label = 3;
                        case 3: return [4 /*yield*/, setTypeForLocation(loc, result)];
                        case 4:
                            _a.sent();
                            return [2 /*return*/, result];
                    }
                });
            });
        }
        var canonizer, typeForCanonicalRef;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    canonizer = new Canonizer();
                    typeForCanonicalRef = immutable_1.Map();
                    return [4 /*yield*/, Support_1.forEachSync(references, function (topLevelRef, topLevelName) { return __awaiter(_this, void 0, void 0, function () {
                            var _a, target, loc, t;
                            return __generator(this, function (_b) {
                                switch (_b.label) {
                                    case 0: return [4 /*yield*/, resolveVirtualRef(undefined, topLevelRef)];
                                    case 1:
                                        _a = _b.sent(), target = _a[0], loc = _a[1];
                                        return [4 /*yield*/, toType(target, loc, TypeNames_2.makeNamesTypeAttributes(topLevelName, false))];
                                    case 2:
                                        t = _b.sent();
                                        typeBuilder.addTopLevel(topLevelName, t);
                                        return [2 /*return*/];
                                }
                            });
                        }); })];
                case 1:
                    _a.sent();
                    return [2 /*return*/];
            }
        });
    });
}
exports.addTypesInSchema = addTypesInSchema;
function definitionRefsInSchema(store, address) {
    return __awaiter(this, void 0, void 0, function () {
        var ref, rootSchema, schema, definitions, definitionsRef;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    ref = Ref.parse(address);
                    return [4 /*yield*/, store.get(ref.address)];
                case 1:
                    rootSchema = _a.sent();
                    schema = ref.lookupRef(rootSchema);
                    if (typeof schema !== "object")
                        return [2 /*return*/, immutable_1.Map()];
                    definitions = schema.definitions;
                    if (typeof definitions !== "object")
                        return [2 /*return*/, immutable_1.Map()];
                    definitionsRef = ref.push("definitions");
                    return [2 /*return*/, immutable_1.Map(Object.getOwnPropertyNames(definitions).map(function (name) {
                            return [name, definitionsRef.push(name)];
                        }))];
            }
        });
    });
}
exports.definitionRefsInSchema = definitionRefsInSchema;
