export interface OptionDefinition {
    name: string;
    type: StringConstructor | BooleanConstructor;
    renderer?: boolean;
    alias?: string;
    multiple?: boolean;
    defaultOption?: boolean;
    defaultValue?: any;
    typeLabel?: string;
    description: string;
    legalValues?: string[];
}
export declare abstract class UntypedOption {
    readonly definition: OptionDefinition;
    constructor(definition: OptionDefinition);
    readonly cliDefinition: OptionDefinition;
}
export declare abstract class Option<T> extends UntypedOption {
    getValue(values: {
        [name: string]: any;
    }): T;
}
export declare class BooleanOption extends Option<boolean> {
    private readonly _cliDescription;
    constructor(name: string, description: string, defaultValue: boolean, _cliDescription?: string | undefined);
    readonly cliDefinition: OptionDefinition;
    getValue(values: {
        [name: string]: any;
    }): boolean;
}
export declare class StringOption extends Option<string> {
    constructor(name: string, description: string, typeLabel: string, defaultValue: string);
}
export declare class EnumOption<T> extends Option<T> {
    private readonly _values;
    constructor(name: string, description: string, values: [string, T][]);
    getValue(values: {
        [name: string]: any;
    }): T;
}
