"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Support_1 = require("./Support");
var TypeAttributes_1 = require("./TypeAttributes");
// FIXME: In the case of overlapping prefixes and suffixes we will
// produce a name that includes the overlap twice.  For example, for
// the names "aaa" and "aaaa" we have the common prefix "aaa" and the
// common suffix "aaa", so we will produce the combined name "aaaaaa".
function combineNames(names) {
    var first = names.first();
    if (first === undefined) {
        return Support_1.panic("Named type has no names");
    }
    if (names.count() === 1) {
        return first;
    }
    var prefixLength = first.length;
    var suffixLength = first.length;
    names.rest().forEach(function (n) {
        prefixLength = Math.min(prefixLength, n.length);
        for (var i = 0; i < prefixLength; i++) {
            if (first[i] !== n[i]) {
                prefixLength = i;
                break;
            }
        }
        suffixLength = Math.min(suffixLength, n.length);
        for (var i = 0; i < suffixLength; i++) {
            if (first[first.length - i - 1] !== n[n.length - i - 1]) {
                suffixLength = i;
                break;
            }
        }
    });
    var prefix = prefixLength > 2 ? first.substr(0, prefixLength) : "";
    var suffix = suffixLength > 2 ? first.substr(first.length - suffixLength) : "";
    var combined = prefix + suffix;
    if (combined.length > 2) {
        return combined;
    }
    return first;
}
var TypeNames = /** @class */ (function () {
    function TypeNames(names, _alternativeNames, areInferred) {
        this.names = names;
        this._alternativeNames = _alternativeNames;
        this.areInferred = areInferred;
    }
    TypeNames.prototype.add = function (names) {
        var newNames = this.names;
        var newAreInferred = this.areInferred;
        if (this.areInferred && !names.areInferred) {
            newNames = names.names;
            newAreInferred = false;
        }
        else if (this.areInferred === names.areInferred) {
            newNames = this.names.union(names.names);
        }
        var newAlternativeNames = this._alternativeNames.union(names._alternativeNames);
        return new TypeNames(newNames, newAlternativeNames, newAreInferred);
    };
    TypeNames.prototype.clearInferred = function () {
        var newNames = this.areInferred ? immutable_1.OrderedSet() : this.names;
        return new TypeNames(newNames, immutable_1.OrderedSet(), this.areInferred);
    };
    Object.defineProperty(TypeNames.prototype, "combinedName", {
        get: function () {
            return combineNames(this.names);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TypeNames.prototype, "proposedNames", {
        get: function () {
            return immutable_1.OrderedSet([this.combinedName]).union(this._alternativeNames);
        },
        enumerable: true,
        configurable: true
    });
    TypeNames.prototype.makeInferred = function () {
        if (this.areInferred)
            return this;
        return new TypeNames(this.names, this._alternativeNames, true);
    };
    TypeNames.prototype.singularize = function () {
        return new TypeNames(this.names.map(pluralize.singular), this._alternativeNames.map(pluralize.singular), true);
    };
    TypeNames.prototype.toString = function () {
        var inferred = this.areInferred ? "inferred" : "given";
        return inferred + " " + this.names.join(",") + " (" + this._alternativeNames.join(",") + ")";
    };
    return TypeNames;
}());
exports.TypeNames = TypeNames;
function typeNamesUnion(c) {
    var names = new TypeNames(immutable_1.OrderedSet(), immutable_1.OrderedSet(), true);
    c.forEach(function (n) {
        names = names.add(n);
    });
    return names;
}
exports.typeNamesUnion = typeNamesUnion;
exports.namesTypeAttributeKind = new TypeAttributes_1.TypeAttributeKind("names", function (a, b) { return a.add(b); }, function (a) { return a.makeInferred(); }, function (a) { return a.toString(); });
function modifyTypeNames(attributes, modifier) {
    return exports.namesTypeAttributeKind.modifyInAttributes(attributes, modifier);
}
exports.modifyTypeNames = modifyTypeNames;
function singularizeTypeNames(attributes) {
    return modifyTypeNames(attributes, function (maybeNames) {
        if (maybeNames === undefined)
            return undefined;
        return maybeNames.singularize();
    });
}
exports.singularizeTypeNames = singularizeTypeNames;
function makeNamesTypeAttributes(nameOrNames, areNamesInferred) {
    var typeNames;
    if (typeof nameOrNames === "string") {
        typeNames = new TypeNames(immutable_1.OrderedSet([nameOrNames]), immutable_1.OrderedSet(), Support_1.defined(areNamesInferred));
    }
    else {
        typeNames = nameOrNames;
    }
    return exports.namesTypeAttributeKind.makeAttributes(typeNames);
}
exports.makeNamesTypeAttributes = makeNamesTypeAttributes;
